#version 120
#extension GL_ARB_texture_rectangle : enable

uniform sampler2DRect texEntry;
uniform sampler2DRect texExit;
uniform sampler3D texVolume;
uniform sampler1D texTransfer;
uniform float dz;

const float DELTA = 0.005;

//const float dz = 0.002;
const int maxrange = 347;

vec3 shading(vec3 N, vec3 V, vec3 L, vec3 color)
{
	float ambientIntensity = 0.2;
	float diffuseIntensity = 0.8;
	float speculaIntensity = 1.0;
	float specExponent = 100;

	vec3 lightColor = vec3(1.0, 1.0, 1.0);
	vec3 H = normalize(L + V);

	float diffuseFactor = max(dot(N, L), 0);
	float specularFactor = pow(max(dot(N, H), 0), specExponent);

	vec3 result = vec3(specularFactor, specularFactor, specularFactor);
	result.x = color.x * ambientIntensity + (color.x * diffuseFactor * diffuseIntensity) + lightColor.x * (specularFactor * speculaIntensity);
	result.y = color.y * ambientIntensity + (color.y * diffuseFactor * diffuseIntensity) + lightColor.y * (specularFactor * speculaIntensity);
	result.z = color.z * ambientIntensity + (color.z * diffuseFactor * diffuseIntensity) + lightColor.z * (specularFactor * speculaIntensity);

	return result;
}

void main()
{
	vec4 front = texture2DRect(texEntry, gl_FragCoord.xy);
	if (front.w == 0)
		discard;

	vec4 back = texture2DRect(texExit, gl_FragCoord.xy);
	vec4 direction = (back - front);
	vec4 stepSize = normalize(direction) * dz;
	int maxiter = int(floor(length(direction) / dz));

	vec4 pos = front; // + ((maxiter / 2) * stepSize);
	vec3 sample1 = vec3(0, 0, 0);
	vec3 sample2 = vec3(0, 0, 0);
	vec4 result = vec4(0, 0, 0, 0);
	for(int i = 0; i < maxiter; i++) {
		sample1.x = texture3D(texVolume, pos.xyz - vec3(DELTA, 0, 0)).x;
		sample2.x = texture3D(texVolume, pos.xyz + vec3(DELTA, 0, 0)).x;
		sample1.y = texture3D(texVolume, pos.xyz - vec3(0, DELTA, 0)).x;
		sample2.y = texture3D(texVolume, pos.xyz + vec3(0, DELTA, 0)).x;
		sample1.z = texture3D(texVolume, pos.xyz - vec3(0, 0, DELTA)).x;
		sample2.z = texture3D(texVolume, pos.xyz + vec3(0, 0, DELTA)).x;

		vec3 N = normalize(sample2 - sample1);
		vec3 L = normalize(vec3(0, 1, 1));
		vec3 V = vec3(direction);

		vec4 intensity = texture3D(texVolume, pos.xyz);
		vec4 transf = texture1D(texTransfer, intensity.x);

		vec3 shadedColor = shading(N, V, L, transf.xyz);

		result.xyz += (1.0 - result.w) * transf.w * shadedColor.xyz;
		result.w += (1.0 - result.w) * transf.w;
		
		pos += stepSize;

		if (result.w >= 1.0)
			break;

		if (pos.x < 0 || pos.x > 1 || pos.y < 0 || pos.y > 1 || pos.z < 0 || pos.z > 1)
			break;
	}

	gl_FragColor = result;
}
