#include "Controls.h"

#include "../VolumeRendering.h"

#include <iostream>

using namespace std;

Controls::Controls(QWidget* parent)
	: QWidget(parent)
{
	this->setObjectName("Controls");

	this->setFixedSize(200, 768);
	this->volumeFileNameLabel = new QLabel(this);
	this->volumeFileNameLabel->setFixedSize(180, 20);
	this->volumeFileNameLabel->move(10, 10);
	this->volumeFileNameLabel->setText("Volume:");
	
	this->volumeFileName = new QLineEdit(this);
	this->volumeFileName->setFixedSize(180, 30);
	this->volumeFileName->setText("volumes/skewed_head.dat");
	this->volumeFileName->move(10, 30);

	this->loadVolumeButton = new QPushButton(this);
	this->loadVolumeButton->setFixedSize(180, 25);
	this->loadVolumeButton->move(10, 60);
	this->loadVolumeButton->setText("Load Volume");
	connect(this->loadVolumeButton, SIGNAL(clicked()), this, SLOT(loadVolumePressed()));

	this->renderModeLabel = new QLabel(this);
	this->renderModeLabel->setFixedSize(180, 20);
	this->renderModeLabel->move(10, 90);
	this->renderModeLabel->setText("Rendermode:");

	this->renderModeBox = new QComboBox(this);
	this->renderModeBox->setFixedSize(180, 20);
	this->renderModeBox->move(10, 110);
	this->renderModeBox->addItem("Slicing");
	this->renderModeBox->addItem("Volume");
	connect(this->renderModeBox, SIGNAL(activated(int)), this, SLOT(renderModeSelected(int)));

	this->viewModeLabel = new QLabel(this);
	this->viewModeLabel->setFixedSize(180, 20);
	this->viewModeLabel->move(10, 130);
	this->viewModeLabel->setText("Viewmode:");

	this->viewModeBox = new QComboBox(this);
	this->viewModeBox->setFixedSize(180, 20);
	this->viewModeBox->move(10, 150);
	this->viewModeBox->addItem("Animated");
	this->viewModeBox->addItem("Manually");
	connect(this->viewModeBox, SIGNAL(activated(int)), this, SLOT(viewModeSelected(int)));

	this->axisLabel = new QLabel(this);
	this->axisLabel->setFixedSize(180, 20);
	this->axisLabel->move(10, 170);
	this->axisLabel->setText("Axis:");

	this->xAxisButton = new QRadioButton(this);
	this->xAxisButton->setFixedSize(60, 20);
	this->xAxisButton->move(10, 190);
	this->xAxisButton->setText("X");
	this->xAxisButton->setChecked(true);
	connect(this->xAxisButton, SIGNAL(clicked()), this, SLOT(xAxisSelected()));

	this->yAxisButton = new QRadioButton(this);
	this->yAxisButton->setFixedSize(60, 20);
	this->yAxisButton->move(70, 190);
	this->yAxisButton->setText("Y");
	connect(this->yAxisButton, SIGNAL(clicked()), this, SLOT(yAxisSelected()));

	this->zAxisButton = new QRadioButton(this);
	this->zAxisButton->setFixedSize(60, 20);
	this->zAxisButton->move(130, 190);
	this->zAxisButton->setText("Z");
	connect(this->zAxisButton, SIGNAL(clicked()), this, SLOT(zAxisSelected()));

	this->axisGroup = new QButtonGroup(this);
	this->axisGroup->addButton(this->xAxisButton);
	this->axisGroup->addButton(this->yAxisButton);
	this->axisGroup->addButton(this->zAxisButton);

	this->axisSlider = new QSlider(Qt::Horizontal, this);
	this->axisSlider->setFixedSize(180, 20);
	this->axisSlider->move(10, 210);
	this->axisSlider->setMaximum(100);
	this->axisSlider->setMinimum(1);
	this->axisSlider->setVisible(false);
	connect(this->axisSlider, SIGNAL(sliderMoved(int)), this, SLOT(axisChanged(int)));

	//this->transferFuncLabel = new QLabel(this);
	//this->transferFuncLabel->setFixedSize(180, 20);
	//this->transferFuncLabel->move(10, 230);
	//this->transferFuncLabel->setText("Transferfunction:");
	
	this->cbTf = new QCheckBox(this);
	this->cbTf->setFixedSize(180,20);
	this->cbTf->move(10,230);
	this->cbTf->setText("Transferfunction:");
	connect( this->cbTf, SIGNAL(stateChanged(int)), this, SLOT(switchedUseTf()));

	this->transferFuncBox = new QComboBox(this);
	this->transferFuncBox->setFixedSize(180, 20);
	this->transferFuncBox->move(10, 250);
	connect(this->transferFuncBox, SIGNAL(activated(int)), this, SLOT(transferFuncSelected(int)));

	this->transferEditor = new ShadeWidget(this);
	this->transferEditor->setFixedSize(180, 390);
	this->transferEditor->move(10, 300);


	this->tfName = new QLineEdit(this);
	this->tfName->setFixedSize(140, 30);
	this->tfName->move(10, 270);

	this->newTfButton = new QPushButton(this);
	this->newTfButton->setFixedSize(40, 30);
	this->newTfButton->move(150,270);
	this->newTfButton->setText("New");
	connect(this->newTfButton, SIGNAL(clicked()), this, SLOT(newTransferFuncPressed()));

	this->sampleDistLabel = new QLabel(this);
	this->sampleDistLabel->setFixedSize(180,20);
	this->sampleDistLabel->move(10,690);
	this->sampleDistLabel->setText("Samplepoint distance:");

	this->axisSampleDist = new QSlider(Qt::Horizontal, this);
	this->axisSampleDist->setFixedSize(180,30);
	this->axisSampleDist->move(10,710);
	this->axisSampleDist->setMaximum(20);
	this->axisSampleDist->setMinimum(1);
	this->axisSampleDist->setSingleStep(1);
	this->axisSampleDist->setTickPosition( QSlider::TicksBelow );
	this->axisSampleDist->setTracking(true);
	connect(this->axisSampleDist, SIGNAL(sliderMoved(int)), this, SLOT(sampleDistChanged()));

	this->screenShotButton = new QPushButton(this);
	this->screenShotButton->setFixedSize(180, 30);
	this->screenShotButton->move(10, 740);
	this->screenShotButton->setText("Screenshot");
	connect(this->screenShotButton, SIGNAL(clicked()), this, SIGNAL(screenShotPressed()));
	this->loadTransferFunctions("tf.xml");

	//std::cout << " xxx " << this->transferFunctions[0].points[0]->pos << std::endl;

	this->transferEditor->set_transferfunction(*this->transferFunctions[0]);
}

Controls::~Controls()
{
}

void Controls::initGLFinished()
{
	if (this->transferFunctions.size() == 0)
		return;

	this->transferEditor->set_transferfunction(*this->transferFunctions[0]);

	VolumeRendering::getRenderer().setTransferFunction(*this->transferFunctions[0]);
}


void Controls::switchedUseTf()
{
    VolumeRendering::getRenderer().setUseTf( (bool) this->cbTf->checkState() );
}

void Controls::sampleDistChanged()
{
    VolumeRendering::getRenderer().setSampleDist( (float) this->axisSampleDist->value() / 1000 );
}

void Controls::newTransferFuncPressed()
{
	QString tfName = this->tfName->text();
   
	Renderer::TransferFunction* tf = new Renderer::TransferFunction();
	
	(*tf).width = 4096;

	Renderer::TFPoint tfPointStart;
	Renderer::TFPoint tfPointEnd;
	
	tfPointStart.pos = 0;
	tfPointStart.r = 0;
	tfPointStart.g = 0;
	tfPointStart.b = 0;
	tfPointStart.a = 0;

	tfPointEnd.pos = 4096;
	tfPointEnd.r = 0;
	tfPointEnd.g = 0;
	tfPointEnd.b = 0;
	tfPointEnd.a = 0;
	
	tf->points.push_back(tfPointStart);
	tf->points.push_back(tfPointEnd);

	this->transferFunctions.push_back(tf);

	this->transferFuncBox->addItem(tfName);
}

void Controls::loadVolumePressed()
{
	string fileName = this->volumeFileName->text ().toStdString();

	if (VolumeRendering::getRenderer().loadVolume(fileName) == false) {
		cout << "Failed to Load Volume \"" << fileName << "\"" << endl;
	}
}

void Controls::renderModeSelected(int index)
{
	VolumeRendering::getRenderer().switchRenderMode(index);
}

void Controls::transferFuncSelected(int index)
{
	cout << "selected TF " << index << endl;

	this->transferEditor->set_transferfunction(*this->transferFunctions[index]);

	VolumeRendering::getRenderer().setTransferFunction(*this->transferFunctions[index]);
}

void Controls::viewModeSelected(int index)
{
	if (index == 0) {
		this->axisSlider->setVisible(false);

		VolumeRendering::getRenderer().setAnimate(true);

	} else if (index == 1) {
		this->axisSlider->setVisible(true);

		VolumeRendering::getRenderer().setAnimate(false);
	}
}

void Controls::xAxisSelected()
{
	VolumeRendering::getRenderer().switchAxis(0);
}

void Controls::yAxisSelected()
{
	VolumeRendering::getRenderer().switchAxis(1);
}

void Controls::zAxisSelected()
{
	VolumeRendering::getRenderer().switchAxis(2);
}

void Controls::axisChanged(int value)
{
	float axisCoord = (float) value / (float) this->axisSlider->maximum();
	VolumeRendering::getRenderer().setAxisCoord(axisCoord);
}

void Controls::saveTransferFunctions(const string& fileName)
{

  std::cout << "saving transferfunctions ..." << std::endl;

  QDomDocument doc;
  QFile file(fileName.c_str());
  if (!file.open(QIODevice::WriteOnly | QIODevice::Truncate | QIODevice::Text))
    return;

  QDomElement root = doc.createElement("transferFunctions");
  doc.appendChild( root );

  std::vector<Renderer::TransferFunction*>::iterator tfIter;
  int i = 0;
  
  for(tfIter = transferFunctions.begin(); tfIter != transferFunctions.end(); tfIter++)
  {
    QDomElement tf = doc.createElement("transferFunction");
    tf.setAttribute( "name", this->transferFuncBox->itemText(i) );
    tf.setAttribute( "width", (*tfIter)->width );
    root.appendChild( tf );

    std::vector<Renderer::TFPoint>::iterator pIter;
    
    for(pIter = (*tfIter)->points.begin(); pIter != (*tfIter)->points.end(); pIter++ )
    {
      QDomElement point = doc.createElement("point");
      point.setAttribute( "pos", (*pIter).pos );
      point.setAttribute( "r", (*pIter).r );
      point.setAttribute( "g", (*pIter).g );
      point.setAttribute( "b", (*pIter).b );
      point.setAttribute( "a", (*pIter).a );

      tf.appendChild( point );
    }

    i++;
  }
  
  QTextStream out(&file);
  out << doc.toString();

  file.close();

  std::cout << "saved" << std::endl;
  
}

// loads all transferfunctions from a given xml-file
void Controls::loadTransferFunctions(const string& fileName)
{
	QDomDocument doc;
	QFile file(fileName.c_str());
	if (!file.open(QIODevice::ReadOnly))
		return;

	if (!doc.setContent(&file)) {
		file.close();
		return;
	}

	QDomElement tfsElement = doc.firstChildElement("transferFunctions");
	if (tfsElement.isNull()) {
		cout << "ERROR in Controls::loadTransferFunctions: failed to get transferFunctions-node" << endl;
		file.close();
		return;
	}

	QDomNode tfNode = tfsElement.firstChild();
	while(!tfNode.isNull()) {
		QDomElement e = tfNode.toElement();
		if(!e.isNull()) {
			if (e.tagName() == "transferFunction") {
				QString tfName;
				Renderer::TransferFunction* tf = this->loadTF(tfName, e);
				if (tf) {
					this->transferFunctions.push_back(tf);
					this->transferFuncBox->addItem(tfName);
				}
			}
		}

		tfNode = tfNode.nextSibling();
	}
}

// returns a single TransferFunction* from a DomElement
Renderer::TransferFunction* Controls::loadTF(QString& tfName, QDomElement& e)
{
	Renderer::TransferFunction* tf = new Renderer::TransferFunction();

	tfName = e.attribute("name");
	tf->width = atoi(e.attribute("width").toStdString().c_str());

	cout << "New TF with Name= " << tfName.toStdString() << " and width = " << tf->width << endl;

	for(QDomNode n = e.firstChild(); !n.isNull(); n = n.nextSibling()) {
		QDomElement e = n.toElement();
		if(!e.isNull()) {
			if (e.tagName() == "point") {
				Renderer::TFPoint tfPoint;
				if (this->loadTFPoint(tfPoint, e))
					tf->points.push_back(tfPoint);
			}
		}
	}

	return tf;
}

// returns a single TransferFunction* from a DomElement
bool Controls::loadTFPoint(Renderer::TFPoint& tfPoint, QDomElement& e)
{
	tfPoint.pos = atoi(e.attribute("pos").toStdString().c_str());
	tfPoint.r = atoi(e.attribute("r").toStdString().c_str());
	tfPoint.g = atoi(e.attribute("g").toStdString().c_str());
	tfPoint.b = atoi(e.attribute("b").toStdString().c_str());
	tfPoint.a = atoi(e.attribute("a").toStdString().c_str());

	cout << "new Point with pos = " << tfPoint.pos << " r = " << (int) tfPoint.r << " g = " << (int) tfPoint.g << " b = " << (int) tfPoint.b << " a = "<< (int) tfPoint.a << endl;

	return true;
}
