#ifndef CONTROLS_H
#define CONTROLS_H

#include "Shadewidget.h"
#include "Hoverpoints.h"
#include "../renderer/Renderer.h"

/**
* @class Controls defines the controls and the window to embedd them
*/
class Controls : public QWidget
{
	Q_OBJECT

 public:
	Controls(QWidget* parent = 0);
	virtual ~Controls();

 signals:
	//! @brief emits this signal when the screen-shot-button is pressed
	void screenShotPressed();

 public slots:
	//! @brief reacts on the finished OpenGL initialization
	void initGLFinished();
	//! @brief loads a Volume from the given filename and sets it as active to be rendered
	void saveTransferFunctions(const std::string&);

 private:
	std::vector<Renderer::TransferFunction*> transferFunctions;

	QLabel* volumeFileNameLabel;
	QLineEdit* volumeFileName;
	QPushButton* loadVolumeButton;

	QLabel* renderModeLabel;
	QComboBox* renderModeBox;

	QLabel* transferFuncLabel;
	QComboBox* transferFuncBox;

	QLabel* viewModeLabel;
	QComboBox* viewModeBox;

	QLabel* axisLabel;
	QButtonGroup* axisGroup;
	QRadioButton* xAxisButton;
	QRadioButton* yAxisButton;
	QRadioButton* zAxisButton;

	QSlider* axisSlider;


	QCheckBox* cbTf;
	QLineEdit* tfName;
	
	QPushButton* newTfButton;


	QLabel* sampleDistLabel;
	QSlider* axisSampleDist;

	ShadeWidget* transferEditor;

	QPushButton* screenShotButton;

 private slots:
	void loadVolumePressed();
	void sampleDistChanged();
	void switchedUseTf();

	void renderModeSelected(int);
	void transferFuncSelected(int);
	void viewModeSelected(int);

	void xAxisSelected();
	void yAxisSelected();
	void zAxisSelected();

	void axisChanged(int);



	void newTransferFuncPressed();
	void loadTransferFunctions(const std::string&);
	Renderer::TransferFunction* loadTF(QString&, QDomElement&);
	bool loadTFPoint(Renderer::TFPoint& tfPoint, QDomElement&);

};

#endif
