#include <sys/time.h>

#include <GL/glew.h>
#include <GL/freeglut.h>

#include "Renderer.h"

#include <vector>
#include <iostream>
#include <string>

#include <assert.h>

#include <string.h>
#include <math.h>

using namespace std;

//! @brief checks the framebuffer status
#define CHECK_FRAMEBUFFER_STATUS() \
{\
 GLenum status; \
 status = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT); \
 switch(status) { \
 case GL_FRAMEBUFFER_COMPLETE_EXT: \
   break; \
 case GL_FRAMEBUFFER_UNSUPPORTED_EXT: \
   fprintf(stderr,"framebuffer GL_FRAMEBUFFER_UNSUPPORTED_EXT\n");\
    /* you gotta choose different formats */ \
   assert(0); \
   break; \
 case GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT_EXT: \
   fprintf(stderr,"framebuffer INCOMPLETE_ATTACHMENT\n");\
   break; \
 case GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT_EXT: \
   fprintf(stderr,"framebuffer FRAMEBUFFER_MISSING_ATTACHMENT\n");\
   break; \
 case GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS_EXT: \
   fprintf(stderr,"framebuffer FRAMEBUFFER_DIMENSIONS\n");\
   break; \
  case GL_FRAMEBUFFER_INCOMPLETE_FORMATS_EXT: \
   fprintf(stderr,"framebuffer INCOMPLETE_FORMATS\n");\
   break; \
 case GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER_EXT: \
   fprintf(stderr,"framebuffer INCOMPLETE_DRAW_BUFFER\n");\
   break; \
 case GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER_EXT: \
   fprintf(stderr,"framebuffer INCOMPLETE_READ_BUFFER\n");\
   break; \
 case GL_FRAMEBUFFER_BINDING_EXT: \
   fprintf(stderr,"framebuffer BINDING_EXT\n");\
   break; \
/* 
 case GL_FRAMEBUFFER_STATUS_ERROR_EXT: \
   fprintf(stderr,"framebuffer STATUS_ERROR\n");\
   break; \
*/ \
 default: \
   /* programming error; will fail on all hardware */ \
   assert(0); \
 }\
}

//! @brief prints the shader information log for a shader object
void printShaderInfoLog(GLuint obj)
{
	int infologLength = 0;
	int charsWritten = 0;
	char* infoLog;

	glGetShaderiv(obj, GL_INFO_LOG_LENGTH, &infologLength);

	if (infologLength > 0) {
		infoLog = (char*) malloc(infologLength);
		glGetShaderInfoLog(obj, infologLength, &charsWritten, infoLog);
		if (charsWritten)
			printf("%s\n",infoLog);
		free(infoLog);
	}
}

//! @brief prints the info log of a program object
void printProgramInfoLog(GLuint obj)
{
	int infologLength = 0;
	int charsWritten  = 0;
	char *infoLog;

	glGetProgramiv(obj, GL_INFO_LOG_LENGTH,&infologLength);
	if (infologLength > 0) {
	        infoLog = (char*) malloc(infologLength);
	        glGetProgramInfoLog(obj, infologLength, &charsWritten, infoLog);
		if (charsWritten)
			printf("%s\n",infoLog);
	        free(infoLog);
	}
}

int GetNextPowerOfTwo(const int iNumber)
{
	int iPowerOfTwo = 1;

	while (iPowerOfTwo < iNumber)
		iPowerOfTwo = iPowerOfTwo * 2;

	return iPowerOfTwo;
};

Renderer::Renderer()
{
	this->volume = 0;

	this->animationFlag = true;	// animation activated
	this->renderMode = 0;		// slicing
	
	this->axisIndex = 0;		// x-achsis
	this->axisCoord = 0;
	this->increaseSlice = true;

	this->loopFactor = 0;

	this->width = 0;
	this->height = 0;

	this->transferTexture = 0;
	this->volumeTexture = 0;
	this->frontTexture = 0;
	this->backTexture = 0;
	this->frontFBO = 0;
	this->backFBO = 0;

	memset(this->rotationAngles, 0, sizeof(this->rotationAngles));

	this->slicingProgram = 0;
	this->volumeProgram = 0;

	this->slicingProgramMit = 0;
	this->volumeProgramMit = 0;

	this->volumeSlicingTextureLocation = 0;

	this->transferTextureLocation = 0;
	this->volumeTextureLocation = 0;
	this->frontTextureLocation = 0;
	this->backTextureLocation = 0;
	this->lightVecLocation = 0;
	this->sampleDistLocation = 0;


	this->volumeSlicingTextureLocationMit = 0;

	this->transferTextureLocationMit = 0;
	this->volumeTextureLocationMit = 0;
	this->frontTextureLocationMit = 0;
	this->backTextureLocationMit = 0;
	this->sampleDistLocationMit = 0;

	this->sampleDist = 0.005;
	this->useTransferfunction = false;
}

Renderer::~Renderer()
{
}

void Renderer::setUseTf(bool useTf)
{
    this->useTransferfunction = useTf;
}

bool Renderer::loadVolume(const std::string& fileName)
{
	// delete an old volume if exists
	if (this->volume) {
		delete this->volume;
		
		// need also to delete the 3D texture from the GPU
		if (this->volumeTexture)
			glDeleteTextures(1, &this->volumeTexture);
	}

	// new volume and load
	this->volume = new Volume();
	this->volume->load(fileName);

	cout << "Volume metrics: (" << this->volume->GetWidth() << "/" << this->volume->GetHeight() << "/" << this->volume->GetDepth() << ")" << endl;

	//create texture
	glGenTextures(1, &this->volumeTexture);
	GLenum glError = GL_NO_ERROR;
	glError = glGetError();
	if (glError != GL_NO_ERROR) {
		std::cerr << "Error generating texture: " << gluErrorString(glError) << std::endl;
		
		delete this->volume;
		this->volume = 0;
		
		this->volumeTexture = 0;

		return false;
	}

	//bind texture
	glBindTexture(GL_TEXTURE_3D, this->volumeTexture);
	glError = glGetError();
	if (glError != GL_NO_ERROR) {
		std::cerr << "Could not bind texture: " << gluErrorString(glError) << std::endl;
		
		delete this->volume;
		this->volume = 0;

		this->volumeTexture = 0;

		return false;
	}

	//upload texture
	glTexImage3D(GL_TEXTURE_3D, 0, GL_LUMINANCE, this->volume->GetWidth(), this->volume->GetHeight(), this->volume->GetDepth(), 0, GL_LUMINANCE, GL_FLOAT, NULL);
	glError = glGetError();
	if (glError != GL_NO_ERROR) {
		cout << "glTexImage3D failed with " << gluErrorString(glError) << endl;

		delete this->volume;
		this->volume = 0;

		this->volumeTexture = 0;

		return false;
	}

	// provide data of the texture
	glTexSubImage3D(GL_TEXTURE_3D, 0, 0, 0, 0, this->volume->GetWidth(), this->volume->GetHeight(), this->volume->GetDepth(), GL_LUMINANCE, GL_FLOAT, (void *) this->volume->Get());
	glError = glGetError();
	if (glError != GL_NO_ERROR) {
		cout << "glTexImage3D failed with " << gluErrorString(glError) << endl;

		delete this->volume;
		this->volume = 0;

		this->volumeTexture = 0;

		return false;
	}

	// setting parameters: linear interpolation
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	// need to clamp to edge, otherwise it could lead to artifacts at the back of the volume: repeating of intensities
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

	glError = glGetError();
	if (glError != GL_NO_ERROR) {
		std::cerr << "Could not upload texture: " << gluErrorString(glError) << std::endl;

		delete this->volume;
		this->volume = 0;

		this->volumeTexture = 0;

		return false;
	}

	return true;
}

void Renderer::setAnimate(bool flag)
{
	this->animationFlag = flag;
}

void Renderer::switchAxis(int axis)
{
	if (axis < 0 || axis > 2)
		return;

	this->axisIndex = axis;
	this->axisCoord = 0.0f;
}

void Renderer::switchRenderMode(int mode)
{
	this->renderMode = mode;
}

void Renderer::setAxisCoord(float coord)
{
	if (this->renderMode == 0) {
		this->axisCoord = coord;
	} else {
		this->rotationAngles[this->axisIndex] = 360 * coord;
	}
}

void Renderer::animate()
{
	if (this->animationFlag == false)
		return;

	if (this->renderMode == 0) {
		this->animateSlicing();
	} else {
		this->animateVolume();
	}
}

void Renderer::animateSlicing()
{
	if (this->volume->GetDepth() > 1) {
		// calculate new coordinate of the selected axis
		if (this->increaseSlice) {
			this->axisCoord += (this->loopFactor / 8);
		} else {
			this->axisCoord -= (this->loopFactor / 8);
		}

		// switch when the maxima or minima has been reached to decrease or decrease the axis coordinate
		if (this->axisCoord > 1.f) {
			this->increaseSlice = false;
			this->axisCoord = 1.f;

		} else if (this->axisCoord < 0.f) {
			this->increaseSlice = true;
			this->axisCoord = 0.f;
		}
	}
}

void Renderer::animateVolume()
{
	// calculate new rotation angle
	float factor = 400; // one quarter rotation per second
	this->rotationAngles[this->axisIndex] += (this->loopFactor * (360 / factor));
	if (this->rotationAngles[this->axisIndex] > 360)
		this->rotationAngles[this->axisIndex] = 0.0;
}

void Renderer::initGL()
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glShadeModel(GL_SMOOTH);
	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
	glClearDepth(1.0f);
	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LEQUAL);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	glEnable(GL_CULL_FACE);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	glViewport(0, 0, this->width, this->height);

	// Initialize GLEW
	std::cout << "- Initializing GLEW ..." << std::endl;
	GLenum err = glewInit();
	if (err != GLEW_OK) {
		// glewInit failed, something is seriously wrong
		std::cerr << "Error initializing GLEW: " << glewGetErrorString(err) << std::endl;
		exit(1);
	}

	this->checkGLFeatures();

	this->loadVolume("volumes/skewed_head.dat");

	cout << "Creating FrontFaces FBO..." << endl;
	this->createFBO(&this->frontTexture, &this->frontFBO);
	cout << "Creating FrontFaces FBO finished" << endl;

	cout << "Creating BackFaces FBO..." << endl;
	this->createFBO(&this->backTexture, &this->backFBO);
	cout << "Creating FrontFaces FBO finished" << endl;

	cout << "Creating Slicing Program..." << endl;
	this->slicingProgram = Renderer::createProgram("shaders/vertexShaderSlicing.glsl", "shaders/fragmentShaderSlicing.glsl");
	if (this->slicingProgram) {
		cout << "Creating Slicing-Program OK" << endl;

		// getting uniform locations for slicing program
		this->volumeSlicingTextureLocation = Renderer::queryUniformLoc(this->slicingProgram, "texVolume");
		this->volumeSlicingTransferLocation = Renderer::queryUniformLoc(this->slicingProgram, "texTransfer");
	} else {
		cout << "Creating Slicing-Program Failed" << endl;
	}

	cout << "Creating Slicing Program (Maximum Intensity Projection)..." << endl;
	this->slicingProgramMit = Renderer::createProgram("shaders/vertexShaderSlicing.glsl", "shaders/fragmentShaderSlicingMIT.glsl");
	if (this->slicingProgramMit) {
		cout << "Creating Slicing-Program (MIT) OK" << endl;

		this->volumeSlicingTextureLocationMit = Renderer::queryUniformLoc(this->slicingProgramMit, "texVolume");
		
	} else {
		cout << "Creating Slicing-Program (MIT) Failed" << endl;
	}
	

	cout << "Creating Volume Program..." << endl;
	this->volumeProgram = Renderer::createProgram("shaders/vertexShaderVolume.glsl", "shaders/fragmentShaderVolume.glsl");
	if (this->volumeProgram) {
		cout << "Creating Volume Program OK" << endl;

		// getting uniform locations for volume program
		this->frontTextureLocation = Renderer::queryUniformLoc(this->volumeProgram, "texEntry");
		this->backTextureLocation = Renderer::queryUniformLoc(this->volumeProgram, "texExit");
		this->volumeTextureLocation = Renderer::queryUniformLoc(this->volumeProgram, "texVolume");
		this->transferTextureLocation = Renderer::queryUniformLoc(this->volumeProgram, "texTransfer");
		this->sampleDistLocation = Renderer::queryUniformLoc(this->volumeProgram, "dz");
	} else {
		cout << "Creating Volume Program Failed" << endl;
	}

	cout << "Creating Volume Program (Maximum Intensity Projection)..." << endl;
	this->volumeProgramMit = Renderer::createProgram("shaders/vertexShaderVolume.glsl", "shaders/fragmentShaderVolumeMIT.glsl");
	if (this->volumeProgramMit) {
		cout << "Creating Volume Program OK" << endl;
  
		this->frontTextureLocationMit = Renderer::queryUniformLoc(this->volumeProgramMit, "texEntry");
		this->backTextureLocationMit = Renderer::queryUniformLoc(this->volumeProgramMit, "texExit");
		this->volumeTextureLocationMit = Renderer::queryUniformLoc(this->volumeProgramMit, "texVolume");
		this->transferTextureLocationMit = Renderer::queryUniformLoc(this->volumeProgramMit, "texTransfer");
		this->sampleDistLocationMit = Renderer::queryUniformLoc(this->volumeProgramMit, "dz");
	} else {
		cout << "Creating Volume Program Failed" << endl;
	}
}

void Renderer::setSampleDist(float in_sampleDist)
{	
   this->sampleDist = in_sampleDist;
}

void Renderer::setTransferFunction(TransferFunction& tf)
{
	cout << "setting new TransferFunction..." << endl;

	// allocate memory for the transferfunction texture
	tf.data = (unsigned char*) malloc(4 * tf.width * sizeof(unsigned char));
	memset(tf.data, 0, 4 * tf.width * sizeof(unsigned char));

	// first point is initial point
	TFPoint prevPoint = tf.points[0];

	std::cout << "( " << (int) prevPoint.r << " , " << (int) prevPoint.g << " , " << (int) prevPoint.b << " , " << (int) prevPoint.a << " ) @ " << (int) prevPoint.pos << std::endl;

	// iterate over all points an linearly interpolate between prev and this
	for (unsigned int i = 1; i < tf.points.size(); i++) {
		TFPoint thisPoint = tf.points[i];

		std::cout << "( " << (int) thisPoint.r << " , " << (int) thisPoint.g << " , " << (int) thisPoint.b << " , " << (int) thisPoint.a << " ) @ " << (int) thisPoint.pos << std::endl;
		unsigned int distance = thisPoint.pos - prevPoint.pos;

		float factor = 0;
		unsigned int index = 0;
		for (unsigned int j = 0; j < distance; j++) {
			index = (prevPoint.pos + j) * 4;
			factor = (float) (j + 1) / (float) distance;

			tf.data[index + 0] = prevPoint.r + (thisPoint.r - prevPoint.r) * factor;
			tf.data[index + 1] = prevPoint.g + (thisPoint.g - prevPoint.g) * factor;
			tf.data[index + 2] = prevPoint.b + (thisPoint.b - prevPoint.b) * factor;
			tf.data[index + 3] = prevPoint.a + (thisPoint.a - prevPoint.a) * factor;
		}

		prevPoint = thisPoint;
	}

	// there has been already a transfertexture set, delete it from GPU
	if (this->transferTexture) {
		cout << "Unloading old Transferfunction..." << endl;
		glDeleteTextures(1, &this->transferTexture);
		cout << "Unloading old Transferfunction succeeded" << endl;
	}

	GLenum glError = GL_NO_ERROR;
	glError = glGetError();
	
	// generate new texid
	glGenTextures(1, &this->transferTexture);

	glError = glGetError();
	if (glError != GL_NO_ERROR)
		cout << "glGenTextures for 1D failed with " << gluErrorString(glError) << endl;

	// bind texid to gpu
	glBindTexture(GL_TEXTURE_1D, this->transferTexture);

	glError = glGetError();
	if (glError != GL_NO_ERROR)
		cout << "glBindTexture for 1D failed with " << gluErrorString(glError) << endl;
	
	// set filter params
	glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MAG_FILTER, GL_LINEAR_MIPMAP_LINEAR);
	glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
	glError = glGetError();
	if (glError != GL_NO_ERROR)
		cout << "glTexParameteri for 1D failed with " << gluErrorString(glError) << endl;

	// upload the data for this transfertexture by creating mipmaps for it
	gluBuild1DMipmaps(GL_TEXTURE_1D, GL_RGBA16, tf.width, GL_RGBA, GL_UNSIGNED_BYTE, tf.data);

	tf.texID = this->transferTexture;

	cout << "setting new TransferFunction succeeded" << endl;
}

void Renderer::createFBO(GLuint* texID, GLuint* fboID)
{
	// create fboid
	glGenFramebuffersEXT(1, fboID);
	// create textureid
	glGenTextures(1, texID);
	if (glGetError() != GL_NO_ERROR)
		cout << "glGenTextures failed with " << gluErrorString(glGetError()) << endl;

	// bind the fbo as framebuffer
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, *fboID);
	// bind the texture
	glBindTexture(GL_TEXTURE_RECTANGLE_ARB, *texID);
	// setup format of the texture
	glTexImage2D(GL_TEXTURE_RECTANGLE_ARB, 0, GL_RGBA16, 1024, 768, 0, GL_RGBA, GL_UNSIGNED_BYTE, 0);
	if (glGetError() != GL_NO_ERROR)
		cout << "glTexImage2D failed with " << gluErrorString(glGetError()) << endl;

	glClearColor(0, 0, 0, 0);
	// link texture to fbo
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_RECTANGLE_ARB, *texID, 0);

	CHECK_FRAMEBUFFER_STATUS()

	// unbind framebuffer
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
}

void Renderer::paintGL()
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);

	timeval t;
	gettimeofday(&t, 0);
	long startTicks = (t.tv_sec * 1000) + t.tv_usec;

	glPushMatrix();

	if (this->renderMode == 0) {
		this->renderSlicing();
	} else {
		this->renderVolume();
	}

	glPopMatrix();

	gettimeofday(&t, 0);
	long endTicks = (t.tv_sec * 1000) + t.tv_usec;

	this->loopFactor = (double)(endTicks - startTicks) / (double) 1000;
}

void Renderer::resizeGL(int w, int h)
{
	this->width = w;
	this->height = h;

	glViewport(0, 0, this->width, this->height);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	gluPerspective(45, (float)this->width / this->height, 1, 1000);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}

void Renderer::renderVolume()
{
	// render frontfaces into frontbuffer
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, this->frontFBO);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	this->renderVolumeBox();

	// render backfaces into backbuffer
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, this->backFBO);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glCullFace(GL_FRONT);
	this->renderVolumeBox();
	glCullFace(GL_BACK);

	// disable framebuffer rendering
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
	glBindTexture(GL_TEXTURE_RECTANGLE_ARB, 0);

	// bind our textures needed for the program
	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_RECTANGLE_ARB, this->frontTexture);
	glActiveTexture(GL_TEXTURE1);
	glBindTexture(GL_TEXTURE_RECTANGLE_ARB, this->backTexture);
	glActiveTexture(GL_TEXTURE2);
	glBindTexture(GL_TEXTURE_3D, this->volumeTexture);
	glActiveTexture(GL_TEXTURE3);
	glBindTexture(GL_TEXTURE_1D, this->transferTexture);

	// activate our shader
	if(this->useTransferfunction)
	  glUseProgram(this->volumeProgram);
	else
	  glUseProgram(this->volumeProgramMit); 

	// set the uniforms which correspond to the textureunits
	if(this->useTransferfunction)
	{
	  glUniform1i(this->frontTextureLocation, 0);
	  glUniform1i(this->backTextureLocation, 1);
	  glUniform1i(this->volumeTextureLocation, 2);
	  glUniform1i(this->transferTextureLocation, 3);
	  glUniform1f(this->sampleDistLocation, this->sampleDist);
	}
	else
	{
	  glUniform1i(this->frontTextureLocationMit, 0);
	  glUniform1i(this->backTextureLocationMit, 1);
	  glUniform1i(this->volumeTextureLocationMit, 2);
	  glUniform1i(this->transferTextureLocationMit, 3);
	  glUniform1f(this->sampleDistLocationMit, this->sampleDist);
	}

	// set orthogonal projection for our display plane
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	// need orthogonal projection with the plane matching the width and height of our viewport
	glOrtho(0.0f, this->height, this->height, 0.0f, -1.0f, 1.0f);

	// draw a quad with the size of the viewport
	glBegin(GL_QUADS);
		glTexCoord2f(0.0, 0.0); glVertex2f(0, 0);
		glTexCoord2f(0.0, 1.0); glVertex2f(0, this->height);
		glTexCoord2f(1.0, 1.0); glVertex2f(this->height, this->height);
		glTexCoord2f(1.0, 0.0); glVertex2f(this->height, 0);
	glEnd();

	// reset the projection back to perspective
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	gluPerspective(45, (float)this->width / this->height, 1, 1000);

	glUseProgram(0);
}

void Renderer::renderVolumeBox()
{
	//glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	//glLoadIdentity();

	float maxDistance = this->volume->GetWidth();
	if (this->volume->GetHeight() > maxDistance)
		maxDistance = this->volume->GetHeight();
	if (this->volume->GetDepth() > maxDistance)
		maxDistance = this->volume->GetDepth();
	
	// position of viewer
	//gluPerspective(45, (float)this->width / this->height, 1, 1000);
	//glViewport(0,0,this->width, this->height);
	gluLookAt(0, 0, 1.5 * maxDistance, 0, 0, 0, 0, 1, 0);


	glRotatef(this->rotationAngles[0], 1.0, 0.0, 0.0);
	glRotatef(this->rotationAngles[1], 0.0, 1.0, 0.0);
	glRotatef(this->rotationAngles[2], 0.0, 0.0, 1.0);

	float halfWidth = this->volume->GetWidth() / 2;
	float halfHeight = this->volume->GetHeight() / 2;
	float halfDepth = this->volume->GetDepth() / 2;

// frontfaces
	glBegin(GL_QUADS);
		glColor3f(1, 1, 1); glVertex3f(halfWidth, halfHeight, halfDepth);
		glColor3f(0, 1, 1); glVertex3f(-halfWidth, halfHeight, halfDepth);
		glColor3f(0, 0, 1); glVertex3f(-halfWidth, -halfHeight, halfDepth);
		glColor3f(1, 0, 1); glVertex3f(halfWidth, -halfHeight, halfDepth);
	glEnd();

	glBegin(GL_QUADS);
		glColor3f(1, 1, 1); glVertex3f(halfWidth, halfHeight, halfDepth);
		glColor3f(1, 1, 0); glVertex3f(halfWidth, halfHeight, -halfDepth);
		glColor3f(0, 1, 0); glVertex3f(-halfWidth, halfHeight, -halfDepth);
		glColor3f(0, 1, 1); glVertex3f(-halfWidth, halfHeight, halfDepth);
	glEnd();

	glBegin(GL_QUADS);
		glColor3f(1, 1, 1); glVertex3f(halfWidth, halfHeight, halfDepth);
		glColor3f(1, 0, 1); glVertex3f(halfWidth, -halfHeight, halfDepth);
		glColor3f(1, 0, 0); glVertex3f(halfWidth, -halfHeight, -halfDepth);
		glColor3f(1, 1, 0); glVertex3f(halfWidth, halfHeight, -halfDepth);
	glEnd();

// backfaces
	glBegin(GL_QUADS);
		glColor3f(0, 0, 0); glVertex3f(-halfWidth, -halfHeight, -halfDepth);
		glColor3f(0, 1, 0); glVertex3f(-halfWidth, halfHeight, -halfDepth);
		glColor3f(1, 1, 0); glVertex3f(halfWidth, halfHeight, -halfDepth);
		glColor3f(1, 0, 0); glVertex3f(halfWidth, -halfHeight, -halfDepth);
	glEnd();

	glBegin(GL_QUADS);
		glColor3f(0, 0, 0); glVertex3f(-halfWidth, -halfHeight, -halfDepth);
		glColor3f(0, 0, 1); glVertex3f(-halfWidth, -halfHeight, halfDepth);
		glColor3f(0, 1, 1); glVertex3f(-halfWidth, halfHeight, halfDepth);
		glColor3f(0, 1, 0); glVertex3f(-halfWidth, halfHeight, -halfDepth);
	glEnd();

	glBegin(GL_QUADS);
		glColor3f(0, 0, 0); glVertex3f(-halfWidth, -halfHeight, -halfDepth);
		glColor3f(1, 0, 0); glVertex3f(halfWidth, -halfHeight, -halfDepth);
		glColor3f(1, 0, 1); glVertex3f(halfWidth, -halfHeight, halfDepth);
		glColor3f(0, 0, 1); glVertex3f(-halfWidth, -halfHeight, halfDepth);
	glEnd();

	glPopMatrix();
}

void Renderer::renderSlicing()
{
	// activate the slicing shader to apply the texture (could also happen with fixed-functionality but its faster and more fun ;)
	if(this->useTransferfunction)
	  glUseProgram(this->slicingProgram);
	else
	  glUseProgram(this->slicingProgramMit);

	// bind our textures needed in this program
	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_3D, this->volumeTexture);
	glActiveTexture(GL_TEXTURE1);
	glBindTexture(GL_TEXTURE_1D, this->transferTexture);

	// set the uniform values (which correspond to the textureunits)
	if(this->useTransferfunction)
	 {
	    glUniform1i(this->volumeSlicingTextureLocation, 0);
	    glUniform1i(this->volumeSlicingTransferLocation, 1);
	}
	else
	{
	    glUniform1i(this->volumeSlicingTextureLocationMit, 0);
	}

	// set orthogonal projection for our display plane
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	glOrtho(0.0f, this->width, this->height, 0.0f, -1.0f, 1.0f);

	// reset the modelview matrix
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	float h = 0.0;
	float w = 0.0;
	float centerX = 0.0;
	float centerY = 0.0;

	// stepping through x.axis: width = volumedepth, height = volume-height
	if (this->axisIndex == 0) {
		w = this->volume->GetDepth();
		h = this->volume->GetHeight();
	// slicing through y achsis: width = volumewidth, height = volumedepth
	} else if (this->axisIndex == 1) {
		w = this->volume->GetWidth();
		h = this->volume->GetDepth();
	// slicing through z achsis: width = volumewidth, height = volumeHeight
	} else if (this->axisIndex == 2) {
		w = this->volume->GetWidth();
		h = this->volume->GetHeight();
	}

	// we maximize the width
	if (w >= h) {
		float factor = this->width / w;
		w *= factor;
		h *= factor;

		centerX = 0.0;
		centerY = (this->height / 2) - (h / 2);

	// we maximize the height
	} else {
		float factor = this->height / h;
		w *= factor;
		h *= factor;

		centerX = (this->width / 2) - (w / 2);
		centerY = 0.0;
	}

	// define our rendering plane
	glBegin(GL_QUADS);

	// we slice through the x-axsis
	if (this->axisIndex == 0) {
		glTexCoord3f(this->axisCoord, 1.0f, 1.0f); glVertex2f(centerX, centerY);
		glTexCoord3f(this->axisCoord, 1.0f, 0.0f); glVertex2f(centerX, h + centerY);
		glTexCoord3f(this->axisCoord, 0.0f, 0.0f); glVertex2f(w + centerX, h + centerY);
		glTexCoord3f(this->axisCoord, 0.0f, 1.0f); glVertex2f(w + centerX, centerY);
	// we slice through the y-axsis
	} else if (this->axisIndex == 1) {
		glTexCoord3f(1.0f, this->axisCoord, 1.0f); glVertex2f(centerX, centerY);
		glTexCoord3f(1.0f, this->axisCoord, 0.0f); glVertex2f(centerX, h + centerY);
		glTexCoord3f(0.0f, this->axisCoord, 0.0f); glVertex2f(w + centerX, h + centerY);
		glTexCoord3f(0.0f, this->axisCoord, 1.0f); glVertex2f(w + centerX, centerY);
	// we slice through the z-axsis
	} else if (this->axisIndex == 2) {
		glTexCoord3f(1.0f, 1.0f, this->axisCoord); glVertex2f(centerX, centerY);
		glTexCoord3f(1.0f, 0.0f, this->axisCoord); glVertex2f(centerX, h + centerY);
		glTexCoord3f(0.0f, 0.0f, this->axisCoord); glVertex2f(w + centerX, h + centerY);
		glTexCoord3f(0.0f, 1.0f, this->axisCoord); glVertex2f(w + centerX, centerY);
	}

	glEnd();

	// reset the projection back to perspective
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	gluPerspective(45, (float)this->width / this->height, 1, 1000);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	glUseProgram(0);
}

GLuint Renderer::createProgram(const string& vertexSourceFile, const string& fragmentSourceFile)
{
	GLint status;
	GLuint program = 0;
	GLuint vertexShader = 0;
	GLuint fragmentShader = 0;

	string vertexSourceStr;
	string fragmentSourceStr;

	if (Renderer::readShaderSource(vertexSourceFile, vertexSourceStr) == false)
		return 0;
	
	if (Renderer::readShaderSource(fragmentSourceFile, fragmentSourceStr) == false)
		return 0;
	
	const char* vertexSource = vertexSourceStr.c_str();
	const char* fragmentSource = fragmentSourceStr.c_str();

	vertexShader = glCreateShader(GL_VERTEX_SHADER);
	if (vertexShader == 0) {
		cout << "glCreateShader for GL_VERTEX_SHADER \"" << vertexSourceFile << "\" failed with " << gluErrorString(glGetError()) << endl;
		return 0;
	}
	fragmentShader = glCreateShader(GL_FRAGMENT_SHADER);
	if (fragmentShader == 0) {
		cout << "glCreateShader for GL_FRAGMENT_SHADER \"" << fragmentSourceFile << "\" failed with " << gluErrorString(glGetError()) << endl;
		return 0;
	}

	glShaderSource(vertexShader, 1, (const GLchar**) &vertexSource, NULL);
	if (glGetError() != GL_NO_ERROR) {
		cout << "glShaderSource for GL_VERTEX_SHADER \"" << vertexSourceFile << "\" failed with " << gluErrorString(glGetError()) << endl;
		return 0;
	}
	glShaderSource(fragmentShader, 1, (const GLchar**) &fragmentSource, NULL);
	if (glGetError() != GL_NO_ERROR) {
		cout << "glShaderSource for GL_FRAGMENT_SHADER \"" << fragmentSourceFile << "\" failed with " << gluErrorString(glGetError()) << endl;
		return 0;
	}

	glCompileShader(vertexShader);
	if (glGetError() != GL_NO_ERROR)
		cout << "glCompileShader for GL_VERTEX_SHADER \"" << vertexSourceFile << "\" failed with " << gluErrorString(glGetError()) << endl;
	glGetShaderiv(vertexShader, GL_COMPILE_STATUS, &status);
	printShaderInfoLog(vertexShader);
	if (!status) {
		cout << "Failed compiling GL_VERTEX_SHADER \"" << vertexSourceFile << "\"" << endl;
		return 0;
	}

	glCompileShader(fragmentShader);
	if (glGetError() != GL_NO_ERROR)
		cout << "glCompileShader for GL_FRAGMENT_SHADER \"" << fragmentSourceFile << "\" failed with " << gluErrorString(glGetError()) << endl;
	glGetShaderiv(fragmentShader, GL_COMPILE_STATUS, &status);
	printShaderInfoLog(fragmentShader);
	if (!status) {
		cout << "Failed compiling GL_FRAGMENT_SHADER \"" << fragmentSourceFile << "\"" << endl;
		return 0;
	}

	program = glCreateProgram();
	if (program == 0) {
		if (glGetError() != GL_NO_ERROR)
			cout << "glCreateProgram failed with " << gluErrorString(glGetError()) << endl;

		return 0;
	}

	glAttachShader(program, vertexShader);
	if (glGetError() != GL_NO_ERROR)
		cout << "glCompileShader for GL_VERTEX_SHADER \"" << vertexSourceFile << "\" failed with " << gluErrorString(glGetError()) << endl;
	glAttachShader(program, fragmentShader);
	if (glGetError() != GL_NO_ERROR)
		cout << "glCompileShader for GL_FRAGMENT_SHADER \"" << fragmentSourceFile << "\" failed with " << gluErrorString(glGetError()) << endl;

	glLinkProgram(program);
	if (glGetError() != GL_NO_ERROR)
		cout << "glLinkProgramARB failed with " << gluErrorString(glGetError()) << endl;
	glGetProgramiv(program, GL_LINK_STATUS, &status);
	printProgramInfoLog(program);
	if (!status) {
		cout << "ERROR linking of program \"" << vertexSourceFile << "\" and \"" << fragmentSourceFile << "\" failed" << endl;
		return 0;
	}

	return program;
}

bool Renderer::readShaderSource(const string& file, string& shaderSource)
{
	string fullFileName = file;
	
	FILE* shaderSourceFile = fopen(fullFileName.c_str(), "r");
	if (shaderSourceFile == 0) {
		cout << "ERROR ... couldn't open Shadersource-File " << fullFileName << endl;
		return false;
	}
	
	char c;
	while ((c = fgetc(shaderSourceFile)) != EOF)
		shaderSource += c;
	
	return true;
}

GLint Renderer::queryUniformLoc(GLint prog, const GLchar* name)
{
	GLint location = 0;
	
	location = glGetUniformLocation(prog, name);
	if (location == -1) {
		cout << "Coulnd't get Uniform Location for name \"" << name << "\". OpenGL-Error: " << gluErrorString(glGetError())  << endl;
		return -1;
	}

	return location;
}

void Renderer::checkGLFeatures()
{
	if (!GLEW_VERSION_2_0) {
		cout << "GLEW_VERSION_2_0 				FAILED" << endl;
		exit(1);
	} else {
		cout << "GLEW_VERSION_2_0 				OK" << endl;
	}

	if (!GLEW_ARB_vertex_program) {
		cout << "GLEW_ARB_vertex_program 			FAILED" << endl;
		exit(1);
	} else {
		cout << "GLEW_ARB_vertex_program 			OK" << endl;
	}

	if (!GLEW_ARB_fragment_program) {
		cout << "GLEW_ARB_fragment_program 			FAILED" << endl;
		exit(1);
	} else {
		cout << "GLEW_ARB_fragment_program 			OK" << endl;
	}

	if (!GLEW_ARB_texture_non_power_of_two) {
		cout << "GLEW_ARB_texture_non_power_of_two		FAILED" << endl;
		exit(1);
	} else {
		cout << "GLEW_ARB_texture_non_power_of_two 		OK" << endl;
	}
}
